/*
 * nghttp2 - HTTP/2 C Library
 *
 * Copyright (c) 2015 Tatsuhiro Tsujikawa
 *
 * Permission is hereby granted, free of charge, to any person obtaining
 * a copy of this software and associated documentation files (the
 * "Software"), to deal in the Software without restriction, including
 * without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to
 * permit persons to whom the Software is furnished to do so, subject to
 * the following conditions:
 *
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE
 * LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 * OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION
 * WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */
#ifndef SHRPX_MEMCACHED_REQUEST_H
#define SHRPX_MEMCACHED_REQUEST_H

#include "shrpx.h"

#include <string>
#include <vector>
#include <memory>

#include "shrpx_memcached_result.h"

namespace shrpx {

enum class MemcachedOp : uint8_t {
  GET = 0x00,
  ADD = 0x02,
};

struct MemcachedRequest;

using MemcachedResultCallback =
  std::function<void(MemcachedRequest *req, MemcachedResult res)>;

struct MemcachedRequest {
  std::string key;
  std::vector<uint8_t> value;
  MemcachedResultCallback cb;
  uint32_t expiry;
  MemcachedOp op;
  bool canceled;
};

} // namespace shrpx

#endif // !defined(SHRPX_MEMCACHED_REQUEST_H)
